@@ -9,9 +9,28 @@ import shutil
 from devpi_common.validation import normalize_name
 from devpi_common.metadata import BasenameMeta
 from devpi_common.types import parse_hash_spec
+from devpi_common.url import URL
 from devpi_server import __version__ as server_version
 from devpi_server.model import is_valid_name
+from devpi_server.model import get_stage_customizer_classes
+from .config import MyArgumentParser
+from .config import add_configfile_option
+from .config import add_export_options
+from .config import add_hard_links_option
+from .config import add_help_option
+from .config import add_import_options
+from .config import add_init_options
+from .config import add_storage_options
+from .config import parseoptions, get_pluginmanager
+from .fileutil import BytesForHardlink
+from .log import configure_cli_logging
+from .main import DATABASE_VERSION
+from .main import Fatal
+from .main import check_python_version
 from .main import fatal
+from .main import init_default_indexes
+from .main import set_state_version
+from .main import xom_from_config
 from .readonly import get_mutable_deepcopy, ReadonlyView
 
 
@@ -46,6 +65,40 @@ def do_export(path, xom):
         dumper.dump_all(path)
     return 0
 
+
+def export(pluginmanager=None, argv=None):
+    """ devpi-export command line entry point. """
+    if argv is None:
+        argv = sys.argv
+    else:
+        # for tests
+        argv = [str(x) for x in argv]
+    check_python_version()
+    if pluginmanager is None:
+        pluginmanager = get_pluginmanager()
+    try:
+        parser = MyArgumentParser(
+            description="Export the data of a devpi-server instance.",
+            add_help=False)
+        add_help_option(parser, pluginmanager)
+        add_configfile_option(parser, pluginmanager)
+        add_storage_options(parser, pluginmanager)
+        add_export_options(parser, pluginmanager)
+        add_hard_links_option(parser, pluginmanager)
+        parser.add_argument("directory")
+        config = parseoptions(pluginmanager, argv, parser=parser)
+        configure_cli_logging(config.args)
+        if not config.path_nodeinfo.exists():
+            fatal("The path '%s' contains no devpi-server data, use devpi-init to initialize." % config.serverdir)
+        xom = xom_from_config(config)
+        do_export(config.args.directory, xom)
+        return 0
+    except Fatal as e:
+        tw = py.io.TerminalWriter(sys.stderr)
+        tw.line("fatal: %s" % e.args[0], red=True)
+        return 1
+
+
 def do_import(path, xom):
     logging.basicConfig(level=logging.INFO, format='%(message)s')
     path = py.path.local(path)
@@ -71,8 +124,52 @@ def do_import(path, xom):
     return 0
 
 
+def import_(pluginmanager=None, argv=None):
+    """ devpi-import command line entry point. """
+    if argv is None:
+        argv = sys.argv
+    else:
+        # for tests
+        argv = [str(x) for x in argv]
+    check_python_version()
+    if pluginmanager is None:
+        pluginmanager = get_pluginmanager()
+    try:
+        parser = MyArgumentParser(
+            description="Import previously exported data into a new devpi-server instance.",
+            add_help=False)
+        add_help_option(parser, pluginmanager)
+        add_configfile_option(parser, pluginmanager)
+        add_storage_options(parser, pluginmanager)
+        add_init_options(parser, pluginmanager)
+        add_import_options(parser, pluginmanager)
+        add_hard_links_option(parser, pluginmanager)
+        parser.add_argument("directory")
+        config = parseoptions(pluginmanager, argv, parser=parser)
+        # BBB set import_ flag on args until the option is removed and the
+        # code adjusted
+        config.args.import_ = True
+        configure_cli_logging(config.args)
+        if config.path_nodeinfo.exists():
+            fatal("The path '%s' already contains devpi-server data." % config.serverdir)
+        sdir = config.serverdir
+        if not (sdir.exists() and len(sdir.listdir()) >= 2):
+            set_state_version(config, DATABASE_VERSION)
+        xom = xom_from_config(config)
+        if config.args.wait_for_events:
+            xom.thread_pool.start_one(xom.keyfs.notifier)
+        init_default_indexes(xom)
+        do_import(config.args.directory, xom)
+        return 0
+    except Fatal as e:
+        tw = py.io.TerminalWriter(sys.stderr)
+        tw.line("fatal: %s" % e.args[0], red=True)
+        return 1
+
+
 class Exporter:
     DUMPVERSION = "2"
+
     def __init__(self, tw, xom):
         self.tw = tw
         self.xom = xom
@@ -110,7 +207,6 @@ class Exporter:
         self.export["dumpversion"] = self.DUMPVERSION
         self.export["pythonversion"] = list(sys.version_info)
         self.export["devpi_server"] = server_version
-        self.export["secret"] = self.config.secret
         self.export["uuid"] = self.xom.config.get_master_uuid()
         for user in self.xom.model.get_userlist():
             userdir = path.join(user.name)
@@ -145,10 +241,16 @@ class IndexDump:
         self.indexmeta = exporter.export_indexes[stage.name] = {}
         self.indexmeta["indexconfig"] = stage.ixconfig
 
-    def dump(self):
+    def should_dump(self):
         if self.stage.ixconfig["type"] == "mirror":
-            projects = []
-        else:
+            if not self.exporter.config.args.include_mirrored_files:
+                return False
+        return True
+
+    def dump(self):
+        projects = []
+        if self.should_dump():
+            self.stage.offline = True
             self.indexmeta["projects"] = {}
             self.indexmeta["files"] = []
             projects = self.stage.list_projects_perstage()
@@ -170,7 +272,9 @@ class IndexDump:
                 self.basedir.ensure(dir=1)
                 self.dump_releasefiles(linkstore)
                 self.dump_toxresults(linkstore)
-                entry = self.stage.get_doczip_entry(vername, version)
+                entry = None
+                if hasattr(self.stage, 'get_doczip_entry'):
+                    entry = self.stage.get_doczip_entry(vername, version)
                 if entry:
                     self.dump_docfile(vername, version, entry)
         self.exporter.completed("index %r" % self.stage.name)
@@ -178,6 +282,8 @@ class IndexDump:
     def dump_releasefiles(self, linkstore):
         for link in linkstore.get_links(rel="releasefile"):
             entry = self.exporter.filestore.get_file_entry(link.entrypath)
+            if not entry.last_modified:
+                continue
             assert entry.file_exists(), entry.relpath
             relpath = self.exporter.copy_file(
                 entry,
@@ -224,6 +330,8 @@ class Importer:
         self.xom = xom
         self.filestore = xom.filestore
         self.tw = tw
+        self.index_customizers = get_stage_customizer_classes(self.xom)
+        self.types_to_skip = set(self.xom.config.args.skip_import_type or [])
 
     def read_json(self, path):
         self.tw.line("reading json: %s" %(path,))
@@ -239,10 +347,8 @@ class Importer:
         total_num_projects = 0
         total_num_files = 0
         for idx_name, idx in self.import_indexes.items():
-            if idx['indexconfig']['type'] == 'mirror':
-                continue
-            num_projects = len(idx['projects'])
-            num_files = len(idx['files'])
+            num_projects = len(idx.get('projects', {}))
+            num_files = len(idx.get('files', []))
             self.tw.line(
                 'Index %s has %d projects and %d files'
                 % (idx_name, num_projects, num_files))
@@ -251,7 +357,8 @@ class Importer:
         self.tw.line('Total number of projects: %d' % total_num_projects)
         self.tw.line('Total number of files: %d' % total_num_files)
 
-    def check_names(self, json_path):
+    def validate(self, json_path):
+        known_types = set(self.index_customizers).union(self.types_to_skip)
         errors = False
         for name in self.import_users:
             if is_valid_name(name):
@@ -261,6 +368,18 @@ class Importer:
                 "Any ascii symbol besides -.@_ is blocked." % name)
             errors = True
         for index in self.import_indexes:
+            config = self.import_indexes[index]['indexconfig']
+            index_type = config['type']
+            if index_type not in known_types:
+                errors = True
+                self.xom.log.error(
+                    "Unknown index type '%s'. "
+                    "Did you forget to install the necessary plugin? "
+                    "You can also skip these with '--skip-import-type %s'.",
+                    index_type, index_type)
+                continue
+            if index_type in self.types_to_skip:
+                continue
             user, name = index.split('/')
             if is_valid_name(name):
                 continue
@@ -270,7 +389,7 @@ class Importer:
             errors = True
         if errors:
             self.xom.log.warn(
-                "You should edit %s manually to fix the above errors." % json_path)
+                "You could also try to edit %s manually to fix the above errors." % json_path)
             raise SystemExit(1)
 
     def iter_projects_normalized(self, projects):
@@ -295,40 +414,52 @@ class Importer:
             self.xom.config.set_uuid(uuid)
         self.import_users = self.import_data["users"]
         self.import_indexes = self.import_data["indexes"]
-        self.xom.config.secret = secret = self.import_data["secret"]
-        self.xom.config.secretfile.write(secret)
         self.display_import_header(path)
-        self.check_names(json_path)
+        self.validate(json_path)
 
         # first create all users
         with self.xom.keyfs.transaction(write=True):
             for username, userconfig in self.import_users.items():
+                user = None
                 if username == "root":
                     user = self.xom.model.get_user(username)
-                else:
+                if user is None:
                     user = self.xom.model.create_user(username, password="")
                 user._set(userconfig)
 
         # memorize index inheritance structure
         tree = IndexTree()
+        indexes = set(self.import_indexes)
         with self.xom.keyfs.transaction(write=False):
             stage = self.xom.model.getstage("root/pypi")
         if stage is not None:
+            indexes.add("root/pypi")
             tree.add("root/pypi")
+        missing_bases = set()
         for stagename, import_index in self.import_indexes.items():
             bases = import_index["indexconfig"].get("bases")
-            tree.add(stagename, bases)
+            if bases is None:
+                tree.add(stagename)
+            else:
+                existing_bases = set(bases).intersection(indexes)
+                missing_bases.update(set(bases) - existing_bases)
+                tree.add(stagename, existing_bases)
+
+        if missing_bases:
+            self.warn(
+                "The following indexes are in bases, but don't exist "
+                "in the import data: %s" % ", ".join(sorted(missing_bases)))
 
         # create stages in inheritance/root-first order
         stages = []
         with self.xom.keyfs.transaction(write=True):
             for stagename in tree.iternames():
-                if stagename == "root/pypi":
-                    stage = self.xom.model.getstage(stagename)
-                    if stage is not None:
-                        continue
+                if stagename == "root/pypi" and stagename not in self.import_indexes:
+                    continue
                 import_index = self.import_indexes[stagename]
                 indexconfig = dict(import_index["indexconfig"])
+                if indexconfig['type'] in self.types_to_skip:
+                    continue
                 if 'uploadtrigger_jenkins' in indexconfig:
                     if not indexconfig['uploadtrigger_jenkins']:
                         # remove if not set, so if the trigger was never
@@ -342,23 +473,34 @@ class Importer:
                 user, index = stagename.split("/")
                 user = self.xom.model.get_user(user)
                 # due to possible circles we create without bases first
-                bases = indexconfig.pop('bases')
-                stage = user.create_stage(index, **indexconfig)
+                # BBB older versions of devpi had bases for mirror indices,
+                # newer versions don't. To support exports from both we
+                # have the default None value
+                bases = indexconfig.pop('bases', None)
+                stage = None
+                if stagename == "root/pypi":
+                    stage = self.xom.model.getstage(stagename)
+                    if stage is not None:
+                        stage.modify(**indexconfig)
+                    elif self.xom.config.args.no_root_pypi:
+                        continue
+                if stage is None:
+                    stage = user.create_stage(index, **indexconfig)
                 if "bases" in import_index["indexconfig"]:
-                    indexconfig = stage.ixconfig
-                    indexconfig["bases"] = bases
-                    stage.modify(**indexconfig)
+                    # we are changing bases directly to allow import with
+                    # removed bases without changing the data from the export
+                    with stage.user.key.update() as userconfig:
+                        indexconfig = userconfig['indexes'][stage.index]
+                        indexconfig["bases"] = tuple(bases)
                 stages.append(stage)
         del tree
 
         # create projects and releasefiles for each index
         for stage in stages:
-            if stage.ixconfig["type"] == "mirror":
-                continue
             imported_files = set()
             import_index = self.import_indexes[stage.name]
-            projects = import_index["projects"]
-            files = import_index["files"]
+            projects = import_index.get("projects", {})
+            files = import_index.get("files", [])
             for project, versions in self.iter_projects_normalized(projects):
                 with self.xom.keyfs.transaction(write=True):
                     for version, versiondata in versions.items():
@@ -372,13 +514,16 @@ class Importer:
                                       "version, setting derived %r" %
                                       (name, version))
                             versiondata["version"] = version
-                        stage.set_versiondata(versiondata)
+                        if hasattr(stage, 'set_versiondata'):
+                            stage.set_versiondata(versiondata)
+                        else:
+                            stage.add_project_name(versiondata["name"])
 
                     # import release files
                     for filedesc in files:
                         if normalize_name(filedesc["projectname"]) == normalize_name(project):
                             imported_files.add(filedesc["relpath"])
-                            self.import_filedesc(stage, filedesc)
+                            self.import_filedesc(stage, filedesc, versions)
             missing = set(x["relpath"] for x in files) - imported_files
             if missing:
                 fatal(
@@ -400,12 +545,16 @@ class Importer:
         self.tw.line("wait_for_events: importing finished"
                      "; latest_serial = %s" % latest_serial)
 
-    def import_filedesc(self, stage, filedesc):
-        assert stage.ixconfig["type"] != "mirror"
+    def import_filedesc(self, stage, filedesc, versions):
         rel = filedesc["relpath"]
         project = filedesc["projectname"]
         p = self.import_rootdir.join(rel)
         assert p.check(), p
+        data = p.read("rb")
+        if self.xom.config.args.hard_links:
+            # wrap the data for additional attribute
+            data = BytesForHardlink(data)
+            data.devpi_srcpath = p.strpath
         if filedesc["type"] == "releasefile":
             mapping = filedesc["entrymapping"]
             if self.dumpversion == "1":
@@ -414,15 +563,33 @@ class Importer:
             else:
                 version = filedesc["version"]
 
-            link = stage.store_releasefile(project, version,
-                                           p.basename, p.read("rb"),
-                                           last_modified=mapping["last_modified"])
+            if hasattr(stage, 'store_releasefile'):
+                link = stage.store_releasefile(
+                    project, version,
+                    p.basename, data,
+                    last_modified=mapping["last_modified"])
+                entry = link.entry
+            else:
+                link = None
+                url = URL(mapping['url']).replace(fragment=mapping['hash_spec'])
+                entry = self.xom.filestore.maplink(
+                    url, stage.username, stage.index, project)
+                entry.file_set_content(data, mapping["last_modified"])
+                (_, links_with_require_python, serial) = stage._load_cache_links(project)
+                if links_with_require_python is None:
+                    links_with_require_python = []
+                links = [(url.basename, entry.relpath)]
+                requires_python = [versions[version].get('requires_python')]
+                for key, href, require_python in links_with_require_python:
+                    links.append((key, href))
+                    requires_python.append(require_python)
+                stage._save_cache_links(project, links, requires_python, serial)
             # devpi-server-2.1 exported with md5 checksums
             if "md5" in mapping:
                 assert "hash_spec" not in mapping
                 mapping["hash_spec"] = "md5=" + mapping["md5"]
             hash_algo, hash_value = parse_hash_spec(mapping["hash_spec"])
-            digest = hash_algo(link.entry.file_get_content()).hexdigest()
+            digest = hash_algo(entry.file_get_content()).hexdigest()
             if digest != hash_value:
                 fatal("File %s has bad checksum %s, expected %s" % (
                       p, digest, hash_value))
@@ -430,7 +597,7 @@ class Importer:
             # determined here but in store_releasefile/store_doczip/store_toxresult etc
         elif filedesc["type"] == "doczip":
             version = filedesc["version"]
-            link = stage.store_doczip(project, version, p.read("rb"))
+            link = stage.store_doczip(project, version, data)
         elif filedesc["type"] == "toxresult":
             linkstore = stage.get_linkstore_perstage(filedesc["projectname"],
                                            filedesc["version"])
@@ -438,14 +605,15 @@ class Importer:
             # it might use a different checksum
             basename = posixpath.basename(filedesc["for_entrypath"])
             link, = linkstore.get_links(basename=basename)
-            link = stage.store_toxresult(link, json.loads(p.read("rb").decode("utf8")))
+            link = stage.store_toxresult(link, json.loads(data.decode("utf8")))
         else:
             fatal("unknown file type: %s" % (type,))
-        history_log = filedesc.get('log')
-        if history_log is None:
-            link.add_log('upload', '<import>', dst=stage.name)
-        else:
-            link.add_logs(history_log)
+        if link is not None:
+            history_log = filedesc.get('log')
+            if history_log is None:
+                link.add_log('upload', '<import>', dst=stage.name)
+            else:
+                link.add_logs(history_log)
 
 
 class IndexTree:
