@@ -2,6 +2,7 @@ import os
 import json
 import contextlib
 import time
+from functools import partial
 from pyramid.httpexceptions import HTTPNotFound, HTTPAccepted, HTTPBadRequest
 from pyramid.httpexceptions import HTTPForbidden
 from pyramid.view import view_config
@@ -10,9 +11,9 @@ from devpi_common.validation import normalize_name
 from webob.headers import EnvironHeaders, ResponseHeaders
 
 from . import mythread
-from .fileutil import loads, rename
+from .fileutil import BytesForHardlink, dumps, loads, rename
 from .log import thread_push_log, threadlog
-from .views import is_mutating_http_method, H_MASTER_UUID, make_uuid_headers
+from .views import H_MASTER_UUID, make_uuid_headers
 from .model import UpstreamError
 
 H_REPLICA_UUID = str("X-DEVPI-REPLICA-UUID")
@@ -21,10 +22,15 @@ H_REPLICA_FILEREPL = str("X-DEVPI-REPLICA-FILEREPL")
 H_EXPECTED_MASTER_ID = str("X-DEVPI-EXPECTED-MASTER-ID")
 
 MAX_REPLICA_BLOCK_TIME = 30.0
+REPLICA_REQUEST_TIMEOUT = MAX_REPLICA_BLOCK_TIME * 1.25
+REPLICA_MULTIPLE_TIMEOUT = REPLICA_REQUEST_TIMEOUT / 2
+MAX_REPLICA_CHANGES_SIZE = 5 * 1024 * 1024
 
 
 class MasterChangelogRequest:
     MAX_REPLICA_BLOCK_TIME = MAX_REPLICA_BLOCK_TIME
+    MAX_REPLICA_CHANGES_SIZE = MAX_REPLICA_CHANGES_SIZE
+    REPLICA_MULTIPLE_TIMEOUT = REPLICA_MULTIPLE_TIMEOUT
 
     def __init__(self, request):
         self.request = request
@@ -40,7 +46,7 @@ class MasterChangelogRequest:
                 "remote-ip": self.request.get_remote_ip(),
                 # the replica always polls its own serial+1
                 # and we want to show where the replica serial is at
-                "serial": int(serial)-1,
+                "serial": int(serial) - 1,
                 "in-request": True,
                 "last-request": time.time(),
                 "outside-url": headers.get(H_REPLICA_OUTSIDE_URL),
@@ -53,19 +59,7 @@ class MasterChangelogRequest:
         else:  # just a regular request
             yield
 
-    @view_config(route_name="/+changelog/{serial}")
-    def get_changes(self):
-        # this method is called from all replica servers
-        # and either returns changelog entry content for {serial} or,
-        # if it points to the "next" serial, will block and wait
-        # until that serial is committed.  However, after
-        # MAX_REPLICA_BLOCK_TIME, we return 202 Accepted to indicate
-        # the replica should try again.  The latter has two benefits:
-        # - nginx' timeout would otherwise return 504 (Gateway Timeout)
-        # - if the replica is not waiting anymore we would otherwise
-        #   never time out here, leading to more and more threads
-        # if no commits happen.
-
+    def verify_master(self):
         if not self.xom.is_master():
             raise HTTPForbidden("Replication protocol disabled")
         expected_uuid = self.request.headers.get(H_EXPECTED_MASTER_ID, None)
@@ -83,18 +77,28 @@ class MasterChangelogRequest:
             raise HTTPBadRequest("expected %s as master_uuid, replica sent %s" %
                                  (master_uuid, expected_uuid))
 
-        serial = self.request.matchdict["serial"]
+    @view_config(route_name="/+changelog/{serial}")
+    def get_changes(self):
+        # this method is called from all replica servers
+        # and either returns changelog entry content for {serial} or,
+        # if it points to the "next" serial, will block and wait
+        # until that serial is committed.  However, after
+        # MAX_REPLICA_BLOCK_TIME, we return 202 Accepted to indicate
+        # the replica should try again.  The latter has two benefits:
+        # - nginx' timeout would otherwise return 504 (Gateway Timeout)
+        # - if the replica is not waiting anymore we would otherwise
+        #   never time out here, leading to more and more threads
+        # if no commits happen.
+
+        self.verify_master()
+
+        serial = int(self.request.matchdict["serial"])
 
         with self.update_replica_status(serial):
             keyfs = self.xom.keyfs
-            if serial.lower() == "nop":
-                raw_entry = b""
-            else:
-                try:
-                    serial = int(serial)
-                except ValueError:
-                    raise HTTPNotFound("serial needs to be int")
-                raw_entry = self._wait_for_entry(serial)
+            self._wait_for_serial(serial)
+
+            raw_entry = keyfs.tx.conn.get_raw_changelog_entry(serial)
 
             devpi_serial = keyfs.get_current_serial()
             r = Response(body=raw_entry, status=200, headers={
@@ -103,7 +107,39 @@ class MasterChangelogRequest:
             })
             return r
 
-    def _wait_for_entry(self, serial):
+    @view_config(route_name="/+changelog/{serial}-")
+    def get_multiple_changes(self):
+        self.verify_master()
+
+        start_serial = int(self.request.matchdict["serial"])
+
+        with self.update_replica_status(start_serial):
+            keyfs = self.xom.keyfs
+            self._wait_for_serial(start_serial)
+            devpi_serial = keyfs.get_current_serial()
+            all_changes = []
+            raw_size = 0
+            start_time = time.time()
+            for serial in range(start_serial, devpi_serial + 1):
+                raw_entry = keyfs.tx.conn.get_raw_changelog_entry(serial)
+                raw_size += len(raw_entry)
+                (changes, rel_renames) = loads(raw_entry)
+                all_changes.append((serial, changes))
+                now = time.time()
+                if raw_size > self.MAX_REPLICA_CHANGES_SIZE:
+                    threadlog.debug('Changelog raw size %s' % raw_size)
+                    break
+                if (now - start_time) > (self.REPLICA_MULTIPLE_TIMEOUT):
+                    threadlog.debug('Changelog timeout %s' % raw_size)
+                    break
+            raw_entry = dumps(all_changes)
+            r = Response(body=raw_entry, status=200, headers={
+                str("Content-Type"): str("application/octet-stream"),
+                str("X-DEVPI-SERIAL"): str(devpi_serial),
+            })
+            return r
+
+    def _wait_for_serial(self, serial):
         keyfs = self.xom.keyfs
         next_serial = keyfs.get_next_serial()
         if serial > next_serial:
@@ -114,15 +150,16 @@ class MasterChangelogRequest:
                 raise HTTPAccepted("no new transaction yet",
                     headers={str("X-DEVPI-SERIAL"):
                              str(keyfs.get_current_serial())})
-        return keyfs.tx.conn.get_raw_changelog_entry(serial)
+        return serial
 
 
 class ReplicaThread:
-    REPLICA_REQUEST_TIMEOUT = MAX_REPLICA_BLOCK_TIME * 1.25
+    REPLICA_REQUEST_TIMEOUT = REPLICA_REQUEST_TIMEOUT
     ERROR_SLEEP = 50
 
     def __init__(self, xom):
         self.xom = xom
+        self.master_auth = xom.config.master_auth
         self.master_url = xom.config.master_url
         self._master_serial = None
         self._master_serial_timestamp = None
@@ -133,6 +170,7 @@ class ReplicaThread:
         self.update_from_master_at = None
         # set whenever the master serial and current replication serial match
         self.replica_in_sync_at = None
+        self.session = self.xom.new_http_session("replica")
 
     def get_master_serial(self):
         return self._master_serial
@@ -156,79 +194,106 @@ class ReplicaThread:
         self._master_serial = serial
         self._master_serial_timestamp = now
 
-    def tick(self):
+    def fetch(self, handler, url):
         log = self.log
-        keyfs = self.xom.keyfs
         config = self.xom.config
-        self.thread.exit_if_shutdown()
-        serial = keyfs.get_next_serial()
-        url = self.master_url.joinpath("+changelog", str(serial)).url
         log.info("fetching %s", url)
         uuid, master_uuid = make_uuid_headers(config.nodeinfo)
         assert uuid != master_uuid
         try:
             self.master_contacted_at = time.time()
-            r = self.session.get(url, headers={
-                H_REPLICA_UUID: uuid,
-                H_EXPECTED_MASTER_ID: master_uuid,
-                H_REPLICA_OUTSIDE_URL: config.args.outside_url,
-            }, timeout=self.REPLICA_REQUEST_TIMEOUT)
+            r = self.session.get(
+                url,
+                auth=self.master_auth,
+                headers={
+                    H_REPLICA_UUID: uuid,
+                    H_EXPECTED_MASTER_ID: master_uuid,
+                    H_REPLICA_OUTSIDE_URL: config.args.outside_url},
+                timeout=self.REPLICA_REQUEST_TIMEOUT)
             remote_serial = int(r.headers["X-DEVPI-SERIAL"])
         except Exception as e:
             log.error("error fetching %s: %s", url, str(e))
-        else:
-            # we check that the remote instance
-            # has the same UUID we saw last time
-            master_uuid = config.get_master_uuid()
-            remote_master_uuid = r.headers.get(H_MASTER_UUID)
-            if not remote_master_uuid:
-                # we don't fatally leave the process because
-                # it might just be a temporary misconfiguration
-                # for example of a nginx frontend
-                log.error("remote provides no %r header, running "
-                          "<devpi-server-2.1?"
-                          " headers were: %s", H_MASTER_UUID, r.headers)
-                self.thread.sleep(self.ERROR_SLEEP)
-                return
-            if master_uuid and remote_master_uuid != master_uuid:
-                # we got a master_uuid and it is not the one we
-                # expect, we are replicating for -- it's unlikely this heals
-                # itself.  It's thus better to die and signal we can't operate.
-                log.error("FATAL: master UUID %r does not match "
-                          "expected master UUID %r. EXITTING.",
-                          remote_master_uuid, master_uuid)
-                # force exit of the process
-                os._exit(3)
+            return False
+        # we check that the remote instance
+        # has the same UUID we saw last time
+        master_uuid = config.get_master_uuid()
+        remote_master_uuid = r.headers.get(H_MASTER_UUID)
+        if not remote_master_uuid:
+            # we don't fatally leave the process because
+            # it might just be a temporary misconfiguration
+            # for example of a nginx frontend
+            log.error("remote provides no %r header, running "
+                      "<devpi-server-2.1?"
+                      " headers were: %s", H_MASTER_UUID, r.headers)
+            self.thread.sleep(self.ERROR_SLEEP)
+            return True
+        if master_uuid and remote_master_uuid != master_uuid:
+            # we got a master_uuid and it is not the one we
+            # expect, we are replicating for -- it's unlikely this heals
+            # itself.  It's thus better to die and signal we can't operate.
+            log.error("FATAL: master UUID %r does not match "
+                      "expected master UUID %r. EXITTING.",
+                      remote_master_uuid, master_uuid)
+            # force exit of the process
+            os._exit(3)
 
-            if r.status_code == 200:
-                try:
-                    changes, rel_renames = loads(r.content)
-                    keyfs.import_changes(serial, changes)
-                except Exception:
-                    log.exception("could not process: %s", r.url)
-                else:
-                    # we successfully received data so let's
-                    # record the master_uuid for future consistency checks
-                    if not master_uuid:
-                        self.xom.config.set_master_uuid(remote_master_uuid)
-                    # also record the current master serial for status info
-                    self.update_master_serial(remote_serial)
-                    return
-            elif r.status_code == 202:
-                log.debug("%s: trying again %s\n", r.status_code, url)
+        if r.status_code == 200:
+            try:
+                handler(r)
+            except Exception:
+                log.exception("could not process: %s", r.url)
+            else:
+                # we successfully received data so let's
+                # record the master_uuid for future consistency checks
+                if not master_uuid:
+                    self.xom.config.set_master_uuid(remote_master_uuid)
                 # also record the current master serial for status info
                 self.update_master_serial(remote_serial)
-                return
-            else:
-                log.error("%s: failed fetching %s", r.status_code, url)
-        # we got an error, let's wait a bit
-        self.thread.sleep(5.0)
+                return True
+        elif r.status_code == 202:
+            log.debug("%s: trying again %s\n", r.status_code, url)
+            # also record the current master serial for status info
+            self.update_master_serial(remote_serial)
+            return True
+        else:
+            log.error("%s: failed fetching %s", r.status_code, url)
+        return False
+
+    def handler_single(self, response, serial):
+        changes, rel_renames = loads(response.content)
+        self.xom.keyfs.import_changes(serial, changes)
+
+    def fetch_single(self, serial):
+        url = self.master_url.joinpath("+changelog", str(serial)).url
+        return self.fetch(
+            partial(self.handler_single, serial=serial),
+            url)
+
+    def handler_multi(self, response):
+        all_changes = loads(response.content)
+        for serial, changes in all_changes:
+            self.xom.keyfs.import_changes(serial, changes)
+
+    def fetch_multi(self, serial):
+        url = self.master_url.joinpath("+changelog", "%s-" % serial).url
+        return self.fetch(self.handler_multi, url)
+
+    def tick(self):
+        self.thread.exit_if_shutdown()
+        serial = self.xom.keyfs.get_next_serial()
+        result = self.fetch_multi(serial)
+        if not result:
+            serial = self.xom.keyfs.get_next_serial()
+            # BBB remove with 6.0.0
+            result = self.fetch_single(serial)
+        if not result:
+            # we got an error, let's wait a bit
+            self.thread.sleep(5.0)
 
     def thread_run(self):
         # within a devpi replica server this thread is the only writer
         self.started_at = time.time()
         self.log = thread_push_log("[REP]")
-        self.session = self.xom.new_http_session("replica")
         keyfs = self.xom.keyfs
         errors = ReplicationErrors(self.xom.config.serverdir)
         for key in (keyfs.STAGEFILE, keyfs.PYPIFILE_NOMD5):
@@ -276,17 +341,6 @@ class SimpleLinksChanged:
                     cache_projectnames.add(project)
 
 
-def tween_replica_proxy(handler, registry):
-    xom = registry["xom"]
-    def handle_replica_proxy(request):
-        assert not hasattr(xom.keyfs, "tx"), "no tx should be ongoing"
-        if is_mutating_http_method(request.method):
-            return proxy_write_to_master(xom, request)
-        else:
-            return handler(request)
-    return handle_replica_proxy
-
-
 hop_by_hop = frozenset((
     'connection',
     'keep-alive',
@@ -332,7 +386,8 @@ def proxy_request_to_master(xom, request, stream=False):
                                 data=request.body,
                                 headers=clean_request_headers(request),
                                 stream=stream,
-                                allow_redirects=False)
+                                allow_redirects=False,
+                                timeout=xom.config.args.proxy_timeout)
         except http.Errors as e:
             raise UpstreamError("proxy-write-to-master %s: %s" % (url, e))
 
@@ -364,6 +419,13 @@ def proxy_write_to_master(xom, request):
                     headers=headers)
 
 
+def proxy_view_to_master(context, request):
+    xom = request.registry["xom"]
+    tx = getattr(xom.keyfs, "tx", None)
+    assert getattr(tx, "write", False) is False, "there should be no write transaction"
+    return proxy_write_to_master(xom, request)
+
+
 class ReplicationErrors:
     def __init__(self, serverdir):
         self.errorsfn = serverdir.join(".replicationerrors")
@@ -398,6 +460,28 @@ class ImportFileReplica:
     def __init__(self, xom, errors):
         self.xom = xom
         self.errors = errors
+        self.file_search_path = self.xom.config.args.replica_file_search_path
+        self.use_hard_links = self.xom.config.args.hard_links
+
+    def find_pre_existing_file(self, key, val):
+        if self.file_search_path is None:
+            return
+        if not os.path.exists(self.file_search_path):
+            threadlog.error(
+                "path for existing files doesn't exist: %s",
+                self.file_search_path)
+        path = os.path.join(self.file_search_path, key.relpath)
+        if os.path.exists(path):
+            threadlog.info("checking existing file: %s", path)
+            with open(path, "rb") as f:
+                data = f.read()
+            if self.use_hard_links:
+                # wrap the data for additional attribute
+                data = BytesForHardlink(data)
+                data.devpi_srcpath = path
+            return data
+        else:
+            threadlog.info("path for existing file not found: %s", path)
 
     def __call__(self, fswriter, key, val, back_serial):
         threadlog.debug("ImportFileReplica for %s, %s", key, val)
@@ -415,12 +499,22 @@ class ImportFileReplica:
             # we have a file or there is no remote file
             return
 
+        content = self.find_pre_existing_file(key, val)
+        if content is not None:
+            err = entry.check_checksum(content)
+            if not err:
+                fswriter.conn.io_file_set(entry._storepath, content)
+                return
+            else:
+                threadlog.error(str(err))
+
         threadlog.info("retrieving file from master: %s", relpath)
         url = self.xom.config.master_url.joinpath(relpath).url
         # we perform the request with a special header so that
         # the master can avoid -getting "volatile" links
-        r = self.xom.httpget(url, allow_redirects=True, extra_headers=
-                             {H_REPLICA_FILEREPL: str("YES")})
+        r = self.xom.httpget(
+            url, allow_redirects=True,
+            extra_headers={H_REPLICA_FILEREPL: str("YES")})
         if r.status_code == 410:
             # master indicates Gone for files which were later deleted
             threadlog.warn("ignoring because of later deletion: %s",
