@@ -1,4 +1,5 @@
 from devpi_server.extpypi import PyPIStage
+from devpi_server.extpypi import PyPICustomizer
 import pytest
 
 
@@ -15,6 +16,9 @@ def pytest_addoption(parser):
 def devpiserver_makepypistage():
     def makepypistage(xom):
         from devpi_server.main import _pypi_ixconfig_default
+        # we copy _pypi_ixconfig_default, otherwise the defaults will
+        # be modified during config updates later on
         return PyPIStage(xom, username="root", index="pypi",
-                         ixconfig=_pypi_ixconfig_default)
+                         ixconfig=dict(_pypi_ixconfig_default),
+                         customizer_cls=PyPICustomizer)
     return makepypistage
