@@ -12,6 +12,7 @@ import requests
 import socket
 import sys
 import time
+from .reqmock import reqmock  # noqa
 from bs4 import BeautifulSoup
 from contextlib import closing
 from devpi_server import extpypi
@@ -21,7 +22,6 @@ from devpi_common.url import URL
 from devpi_server.extpypi import PyPIStage
 from devpi_server.log import threadlog, thread_clear_log
 from pyramid.authentication import b64encode
-from pyramid.compat import escape
 from pyramid.httpexceptions import status_map
 
 import hashlib
@@ -30,10 +30,12 @@ try:
 except ImportError:
     from Queue import Queue as BaseQueue
 
-try:
-    import http.server as httpserver
-except ImportError:
-    import BaseHTTPServer as httpserver
+
+@pytest.fixture(scope="session")
+def server_version():
+    from devpi_server import __version__
+    from pkg_resources import parse_version
+    return parse_version(__version__)
 
 
 def make_file_url(basename, content, stagename=None, baseurl="http://localhost/"):
@@ -49,6 +51,7 @@ class TimeoutQueue(BaseQueue):
     def get(self, timeout=2):
         return BaseQueue.get(self, timeout=timeout)
 
+
 log = threadlog
 
 @pytest.fixture(autouse=True)
@@ -93,13 +96,12 @@ def caplog(caplog):
     return caplog
 
 @pytest.fixture
-def gentmp(request):
-    tmpdirhandler = request.config._tmpdirhandler
+def gentmp(request, tmpdir_factory):
     cache = []
     def gentmp(name=None):
         if not cache:
             prefix = re.sub(r"[\W]", "_", request.node.name)
-            basedir = tmpdirhandler.mktemp(prefix, numbered=True)
+            basedir = tmpdir_factory.mktemp(prefix, numbered=True)
             cache.append(basedir)
         else:
             basedir = cache[0]
@@ -194,10 +196,11 @@ def makexom(request, gentmp, httpget, monkeypatch, storage_info):
     def makexom(opts=(), httpget=httpget, plugins=()):
         from devpi_server import auth_basic
         from devpi_server import auth_devpi
+        from devpi_server import model
         plugins = [
             plugin[0] if isinstance(plugin, tuple) else plugin
             for plugin in plugins]
-        for plugin in [auth_basic, auth_devpi, storage_info["_test_plugin"]]:
+        for plugin in [auth_basic, auth_devpi, extpypi, model, storage_info["_test_plugin"]]:
             if plugin not in plugins:
                 plugins.append(plugin)
         pm = get_pluginmanager(load_entrypoints=False)
@@ -230,10 +233,9 @@ def makexom(request, gentmp, httpget, monkeypatch, storage_info):
                     lambda self: set())
             add_pypistage_mocks(monkeypatch, httpget)
         # initialize default indexes
-        from devpi_server.main import set_default_indexes
+        from devpi_server.main import init_default_indexes
         if not xom.config.args.master_url:
-            with xom.keyfs.transaction(write=True):
-                set_default_indexes(xom.model)
+            init_default_indexes(xom)
         if request.node.get_closest_marker("with_replica_thread"):
             from devpi_server.replica import ReplicaThread
             rt = ReplicaThread(xom)
@@ -287,43 +289,22 @@ def makemapp(request, maketestapp, makexom):
     return makemapp
 
 
-def make_simple_pkg_info(name, text="", pkgver=None, hash_type=None,
-                         pypiserial=None, requires_python=None):
-    class ret:
-        hash_spec = ""
-    if requires_python:
-        requires_python = ' data-requires-python="%s"' % escape(requires_python)
-    else:
-        requires_python = ''
-    if pkgver is not None:
-        assert not text
-        if hash_type and "#" not in pkgver:
-            hv = (pkgver + str(pypiserial)).encode("ascii")
-            hash_value = getattr(hashlib, hash_type)(hv).hexdigest()
-            ret.hash_spec = "%s=%s" %(hash_type, hash_value)
-            pkgver += "#" + ret.hash_spec
-        text = '<a href="../../{name}/{pkgver}"{requires_python}>{pkgver}</a>'.format(
-            name=name, pkgver=pkgver, requires_python=requires_python)
-    elif text and "{md5}" in text:
-        text = text.format(md5=getmd5(text))
-    elif text and "{sha256}" in text:
-        text = text.format(sha256=getsha256(text))
-    return ret, text
-
-
 @pytest.fixture
 def httpget(pypiurls):
+    from .simpypi import make_simple_pkg_info
     class MockHTTPGet:
         def __init__(self):
             self.url2response = {}
             self._md5 = hashlib.md5()
 
-        def __call__(self, url, allow_redirects=False, extra_headers=None):
+        def __call__(self, url, allow_redirects=False, extra_headers=None, **kw):
             class mockresponse:
                 def __init__(xself, url):
                     fakeresponse = self.url2response.get(url)
                     if fakeresponse is None:
-                        fakeresponse = dict(status_code = 404)
+                        fakeresponse = dict(
+                            status_code=404,
+                            reason="Not Found")
                     xself.__dict__.update(fakeresponse)
                     if "url" not in fakeresponse:
                         xself.url = url
@@ -496,7 +477,9 @@ class Mapp(MappMixin):
         if user is None:
             user = self.testapp.auth[0]
         r = self.testapp.get("/%s/%s" % (indexname, name), accept="application/json")
-        assert r.status_code == 200
+        assert r.status_code == code
+        if r.status_code >= 300:
+            return
         result = r.json["result"]
         links = set()
         for version in result.values():
@@ -593,7 +576,8 @@ class Mapp(MappMixin):
                                   expect_errors=True)
         assert r.status_code == code
         if code in (200,201):
-            assert r.json["result"]["type"] == indexconfig.get("type", "stage")
+            if isinstance(indexconfig, dict):
+                assert r.json["result"]["type"] == indexconfig.get("type", "stage")
             return r.json["result"]
         if code in (400,):
             return r.json["message"]
@@ -845,7 +829,6 @@ class MyTestApp(TApp):
         assert r.status_code == code
         return r
 
-
     def get_json(self, *args, **kwargs):
         headers = kwargs.setdefault("headers", {})
         headers["Accept"] = "application/json"
@@ -907,69 +890,6 @@ def testapp(request, maketestapp, xom):
     return maketestapp(xom)
 
 
-class SimPyPIRequestHandler(httpserver.BaseHTTPRequestHandler):
-    def do_GET(self):
-        def start_response(status, headers):
-            self.send_response(status)
-            for key, value in headers.items():
-                self.send_header(key, value)
-            self.end_headers()
-
-        simpypi = self.server.simpypi
-        headers = {
-            'X-Simpypi-Method': 'GET'}
-        p = self.path.split('/')
-        if len(p) == 4 and p[0] == '' and p[1] == 'simple' and p[3] == '':
-            # project listing
-            project = simpypi.projects.get(p[2])
-            if project is not None:
-                releases = project['releases']
-                simpypi.add_log(
-                    "do_GET", self.path, "found",
-                    project['title'], "with", list(releases))
-                start_response(200, headers)
-                self.wfile.write(b'\n'.join(releases))
-                return
-        elif p == ['', 'simple', ''] or p == ['', 'simple']:
-            # root listing
-            projects = [
-                '<a href="/simple/%s/">%s</a>' % (k, v['title'])
-                for k, v in simpypi.projects.items()]
-            simpypi.add_log("do_GET", self.path, "found", list(simpypi.projects))
-            start_response(200, headers)
-            self.wfile.write(b'\n'.join(x.encode('utf-8') for x in projects))
-            return
-        elif self.path in simpypi.files:
-            # file serving
-            f = simpypi.files[self.path]
-            content = f['content']
-            if 'length' in f:
-                headers['Content-Length'] = f['length']
-                content = content[:f['length']]
-            start_response(200, headers)
-            simpypi.add_log("do_GET", self.path, "sending")
-            if not f['stream']:
-                self.wfile.write(content)
-                simpypi.add_log("do_GET", self.path, "sent")
-                return
-            else:
-                chunksize = f['chunksize']
-                callback = f.get('callback')
-                for i in range(len(content) // chunksize):
-                    data = content[i * chunksize:(i + 1) * chunksize]
-                    if not data:
-                        break
-                    self.wfile.write(data)
-                    if callback:
-                        callback(i * chunksize)
-                    simpypi.add_log(
-                        "do_GET", self.path,
-                        "streamed %i bytes" % (i * chunksize))
-                return
-        simpypi.add_log("do_GET", self.path, "not found")
-        start_response(404, headers)
-
-
 def get_open_port(host):
     with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
         s.bind((host, 0))
@@ -990,7 +910,7 @@ def wait_for_port(host, port, timeout=60):
         "The port %s on host %s didn't become accessible" % (port, host))
 
 
-@pytest.yield_fixture(scope="module")
+@pytest.yield_fixture(scope="class")
 def server_directory():
     import tempfile
     srvdir = py.path.local(
@@ -1002,20 +922,37 @@ def server_directory():
 @pytest.fixture(scope="module")
 def call_devpi_in_dir():
     # let xproc find the correct executable instead of py.test
+    devpigenconfig = str(py.path.local.sysfind("devpi-gen-config"))
+    devpiimport = str(py.path.local.sysfind("devpi-import"))
+    devpiinit = str(py.path.local.sysfind("devpi-init"))
     devpiserver = str(py.path.local.sysfind("devpi-server"))
 
     def devpi(server_dir, args):
+        from devpi_server.genconfig import genconfig
+        from devpi_server.importexport import import_
+        from devpi_server.init import init
         from devpi_server.main import main
         from _pytest.monkeypatch import MonkeyPatch
         from _pytest.pytester import RunResult
         m = MonkeyPatch()
-        m.setenv("DEVPISERVER_SERVERDIR", server_dir)
-        m.setattr("sys.argv", [devpiserver])
+        m.setenv("DEVPISERVER_SERVERDIR", getattr(server_dir, 'strpath', server_dir))
         cap = py.io.StdCaptureFD()
         cap.startall()
         now = time.time()
+        if args[0] == 'devpi-gen-config':
+            m.setattr("sys.argv", [devpigenconfig])
+            entry_point = genconfig
+        elif args[0] == 'devpi-import':
+            m.setattr("sys.argv", [devpiimport])
+            entry_point = import_
+        elif args[0] == 'devpi-init':
+            m.setattr("sys.argv", [devpiinit])
+            entry_point = init
+        elif args[0] == 'devpi-server':
+            m.setattr("sys.argv", [devpiserver])
+            entry_point = main
         try:
-            main(args)
+            entry_point(argv=args)
         finally:
             m.undo()
             out, err = cap.reset()
@@ -1032,20 +969,35 @@ def master_serverdir(server_directory):
 
 
 @pytest.yield_fixture(scope="class")
-def master_host_port(request, call_devpi_in_dir, master_serverdir):
+def master_host_port(request, call_devpi_in_dir, master_serverdir, server_directory, storage_info):
+    import base64
+    import os
+    secretfile = server_directory.join('testserver.secret')
+    if not secretfile.exists():
+        secretfile.write(base64.b64encode(os.urandom(32)))
     host = 'localhost'
     port = get_open_port(host)
     args = [
         "devpi-server",
-        "--serverdir", master_serverdir.strpath,
         "--role", "master",
+        "--secretfile", secretfile.strpath,
         "--host", host,
         "--port", str(port),
         "--requests-only"]
+    storage_args = [
+        "--serverdir", master_serverdir.strpath]
+    if storage_info["name"] != "sqlite":
+        storage_option = "--storage=%s" % storage_info["name"]
+        _get_test_storage_options = getattr(
+            storage_info["storage"], "_get_test_storage_options", None)
+        if _get_test_storage_options:
+            storage_options = _get_test_storage_options(master_serverdir)
+            storage_option = storage_option + storage_options
+        storage_args.append(storage_option)
     if not master_serverdir.join('.nodeinfo').exists():
         subprocess.check_call(
-            args + ["--init"])
-    p = subprocess.Popen(args)
+            ["devpi-init"] + storage_args)
+    p = subprocess.Popen(args + storage_args)
     try:
         wait_for_port(host, port)
         yield (host, port)
@@ -1060,25 +1012,34 @@ def replica_serverdir(server_directory):
 
 
 @pytest.yield_fixture(scope="class")
-def replica_host_port(request, call_devpi_in_dir, master_host_port, replica_serverdir):
+def replica_host_port(request, call_devpi_in_dir, master_host_port, replica_serverdir, storage_info):
     host = 'localhost'
     port = get_open_port(host)
     args = [
-        "devpi-server", "--start",
-        "--host", host, "--port", str(port),
-        "--master-url", "http://%s:%s" % master_host_port]
+        "devpi-server",
+        "--host", host, "--port", str(port)]
+    storage_args = [
+        "--serverdir", replica_serverdir.strpath]
+    if storage_info["name"] != "sqlite":
+        storage_option = "--storage=%s" % storage_info["name"]
+        _get_test_storage_options = getattr(
+            storage_info["storage"], "_get_test_storage_options", None)
+        if _get_test_storage_options:
+            storage_options = _get_test_storage_options(replica_serverdir)
+            storage_option = storage_option + storage_options
+        storage_args.append(storage_option)
     if not replica_serverdir.join('.nodeinfo').exists():
-        args.append("--init")
-    call_devpi_in_dir(
-        replica_serverdir.strpath,
-        args)
+        subprocess.check_call([
+            "devpi-init",
+            "--role", "replica",
+            "--master-url", "http://%s:%s" % master_host_port] + storage_args)
+    p = subprocess.Popen(args + storage_args)
     try:
         wait_for_port(host, port)
         yield (host, port)
     finally:
-        call_devpi_in_dir(
-            replica_serverdir.strpath,
-            ["devpi-server", "--stop"])
+        p.terminate()
+        p.wait()
 
 
 nginx_conf_content = """
@@ -1110,9 +1071,9 @@ def _nginx_host_port(host, port, call_devpi_in_dir, server_directory):
 
     orig_dir = server_directory.chdir()
     try:
-        args = ["devpi-server", "--gen-config", "--host", host, "--port", str(port)]
+        args = ["devpi-gen-config", "--host", host, "--port", str(port)]
         if not server_directory.join('.nodeinfo').exists():
-            args.append("--init")
+            call_devpi_in_dir(server_directory.strpath, ["devpi-init"])
         call_devpi_in_dir(
             server_directory.strpath,
             args)
@@ -1171,6 +1132,7 @@ def nginx_replica_host_port(replica_host_port, call_devpi_in_dir, server_directo
 
 @pytest.fixture(scope="session")
 def simpypiserver():
+    from .simpypi import httpserver, SimPyPIRequestHandler
     import threading
     host = 'localhost'
     port = get_open_port(host)
@@ -1183,63 +1145,9 @@ def simpypiserver():
     return server
 
 
-class SimPyPI:
-    def __init__(self, address):
-        self.baseurl = "http://%s:%s" % address
-        self.simpleurl = "%s/simple" % self.baseurl
-        self.projects = {}
-        self.files = {}
-        self.clear_log()
-
-    def clear_log(self):
-        self.log = []
-
-    def add_log(self, *args):
-        msg = ' '.join(str(x) for x in args)
-        print(msg, file=sys.stderr)
-        self.log.append(msg)
-
-    def add_project(self, name, title=None):
-        if title is None:
-            title = name
-        self.projects.setdefault(
-            name, dict(
-                title=title,
-                releases=set(),
-                pypiserial=None))
-        return self.projects[name]
-
-    def remove_project(self, name):
-        self.projects.pop(name)
-
-    def add_release(self, name, title=None, text="", pkgver=None, hash_type=None,
-                    pypiserial=None, requires_python=None, **kw):
-        project = self.add_project(name, title=title)
-        ret, text = make_simple_pkg_info(
-            name, text=text, pkgver=pkgver, hash_type=hash_type,
-            pypiserial=pypiserial, requires_python=requires_python)
-        assert text
-        project['releases'].add(text.encode('utf-8'))
-
-    def add_file(self, relpath, content, stream=False, chunksize=1024,
-                 length=None, callback=None):
-        if length is None:
-            length = len(content)
-        info = dict(
-            content=content,
-            stream=stream,
-            chunksize=chunksize,
-            callback=callback)
-        if length is not False:
-            info['length'] = length
-        self.files[relpath] = info
-
-    def remove_file(self, relpath):
-        del self.files[relpath]
-
-
 @pytest.fixture
 def simpypi(simpypiserver):
+    from .simpypi import SimPyPI
     simpypiserver.simpypi = SimPyPI(simpypiserver.server_address)
     return simpypiserver.simpypi
 
@@ -1258,84 +1166,6 @@ def pytest_runtest_setup(item):
         if previousfailed is not None:
             pytest.xfail("previous test failed (%s)" %previousfailed.name)
 
-#
-#  various requests related mocking functionality
-#  (XXX consolidate, release a plugin?)
-#
-from requests.adapters import HTTPAdapter
-from requests.packages.urllib3.response import HTTPResponse
-import fnmatch
-
-
-@pytest.fixture
-def reqmock(monkeypatch):
-    mr = mocked_request()
-    def get_adapter(self, url):
-        return MockAdapter(mr, url)
-    monkeypatch.setattr("requests.sessions.Session.get_adapter", get_adapter)
-    return mr
-
-class MockAdapter:
-    def __init__(self, mock_request, url):
-        self.url = url
-        self.mock_request = mock_request
-
-    def send(self, request, **kwargs):
-        return self.mock_request.process_request(request, kwargs)
-
-
-class mocked_request:
-    def __init__(self):
-        self.url2reply = {}
-
-    def process_request(self, request, kwargs):
-        url = request.url
-        response = self.url2reply.get((url, request.method))
-        if response is None:
-            response = self.url2reply.get((url, None))
-            if response is None:
-                for (name, method), response in self.url2reply.items():
-                    if method is None or method == request.method:
-                        if fnmatch.fnmatch(request.url, name):
-                            break
-                else:
-                    raise Exception("not mocked call to %s" % url)
-        response.add_request(request)
-        r = HTTPAdapter().build_response(request, response)
-        return r
-
-    def mockresponse(self, url, code, method=None, data=None, headers=None,
-                     on_request=None, reason=None):
-        if not url:
-            url = "*"
-        r = ReqReply(code=code, data=data, headers=headers,
-                     on_request=on_request, reason=reason)
-        if method is not None:
-            method = method.upper()
-        self.url2reply[(url, method)] = r
-        return r
-    mock = mockresponse
-
-class ReqReply(HTTPResponse):
-    def __init__(self, code, data, headers, on_request, reason=None):
-        if py.builtin._istext(data):
-            data = data.encode("utf-8")
-        super(ReqReply, self).__init__(body=py.io.BytesIO(data),
-                                       status=code,
-                                       headers=headers,
-                                       reason=reason,
-                                       preload_content=False)
-        self.requests = []
-        self.on_request = on_request
-
-    def add_request(self, request):
-        if self.on_request:
-            self.on_request(request)
-        self.requests.append(request)
-
-#
-#  end requests related mocking functionality
-#
 
 @pytest.fixture
 def gen():
@@ -1354,12 +1184,6 @@ class Gen:
             link += "#md5=%s" % md5
         return URL(link)
 
-def getmd5(s):
-    return hashlib.md5(s.encode("utf8")).hexdigest()
-
-def getsha256(s):
-    return hashlib.sha256(s.encode("utf8")).hexdigest()
-
 
 @pytest.yield_fixture
 def pyramidconfig():
@@ -1383,3 +1207,14 @@ def blank_request():
     def blank_request(*args, **kwargs):
         return Request.blank("/blankpath", *args, **kwargs)
     return blank_request
+
+
+@pytest.fixture(params=[None, "tox38"])
+def tox_result_data(request):
+    from test_devpi_server.example import tox_result_data
+    import copy
+    tox_result_data = copy.deepcopy(tox_result_data)
+    if request.param == "tox38":
+        retcode = int(tox_result_data['testenvs']['py27']['test'][0]['retcode'])
+        tox_result_data['testenvs']['py27']['test'][0]['retcode'] = retcode
+    return tox_result_data
