@@ -1,6 +1,7 @@
 import pytest
 from devpi_server import mythread
 from devpi_server.config import hookimpl
+from devpi_server.config import parseoptions, get_pluginmanager
 from devpi_server.main import Fatal
 from devpi_server.main import XOM
 from devpi_server.main import check_compatible_version
@@ -13,9 +14,19 @@ import os
 def ground_wsgi_run(monkeypatch):
     monkeypatch.setattr(mythread.ThreadPool, "live", lambda *args: 0 / 0)
 
+
 wsgi_run_throws = pytest.mark.usefixtures("ground_wsgi_run")
 
 
+@pytest.fixture
+def config(gentmp):
+    serverdir = gentmp()
+    pluginmanager = get_pluginmanager()
+    return parseoptions(
+        pluginmanager,
+        ["devpi-server", "--serverdir", serverdir.strpath])
+
+
 def test_pkgresources_version_matches_init():
     import pkg_resources
     ver = devpi_server.__version__
@@ -27,36 +38,44 @@ def test_version(capfd):
     assert not err  # not logging output
     assert devpi_server.__version__ in out.strip()
 
-def test_check_compatible_version_earlier(xom, monkeypatch):
-    monkeypatch.setattr(xom, "get_state_version", lambda: "1.0")
+def test_check_compatible_version_earlier(config, monkeypatch):
+    monkeypatch.setattr("devpi_server.main.get_state_version", lambda cfg: "1.0")
     with pytest.raises(Fatal):
-        check_compatible_version(xom)
+        check_compatible_version(config)
 
-def test_check_compatible_version_self(xom):
-    check_compatible_version(xom)
 
-def test_check_compatible_dev_version(xom, monkeypatch):
-    monkeypatch.setattr(xom, "get_state_version", lambda: "2.3.0.dev1")
-    monkeypatch.setattr("devpi_server.main.server_version", "2.3.0.dev2")
-    check_compatible_version(xom)
+@pytest.mark.parametrize("version", ["4", "4.8.1"])
+def test_check_compatible_version(config, version):
+    versionfile = config.serverdir.join(".serverversion")
+    versionfile.write(version)
+    check_compatible_version(config)
+
+
+def test_check_compatible_dev_version(config, monkeypatch):
+    monkeypatch.setattr("devpi_server.main.get_state_version", lambda cfg: "2.3.0.dev1")
+    monkeypatch.setattr("devpi_server.main.DATABASE_VERSION", "2")
+    check_compatible_version(config)
+
+
+def test_check_compatible_minor_version(config, monkeypatch):
+    monkeypatch.setattr("devpi_server.main.get_state_version", lambda cfg: "2.2.0")
+    monkeypatch.setattr("devpi_server.main.DATABASE_VERSION", "2")
+    check_compatible_version(config)
 
-def test_check_compatible_minor_version(xom, monkeypatch):
-    monkeypatch.setattr(xom, "get_state_version", lambda: "2.2.0")
-    monkeypatch.setattr("devpi_server.main.server_version", "2.3.0")
-    check_compatible_version(xom)
 
 @pytest.mark.parametrize("version", ["2.0.1", "2.1.9"])
-def test_check_incompatible_minor_version_raises(xom, monkeypatch, version):
-    monkeypatch.setattr(xom, "get_state_version", lambda: version)
-    monkeypatch.setattr("devpi_server.main.server_version", "2.3.0")
+def test_check_incompatible_minor_version_raises(config, monkeypatch, version):
+    monkeypatch.setattr("devpi_server.main.get_state_version", lambda cfg: version)
     with pytest.raises(Fatal):
-        check_compatible_version(xom)
+        check_compatible_version(config)
 
-def test_check_incompatible_version_raises(xom):
-    versionfile = xom.config.serverdir.join(".serverversion")
+
+def test_check_incompatible_version_raises(config):
+    versionfile = config.serverdir.join(".serverversion")
     versionfile.write("5.0.4")
     with pytest.raises(Fatal):
-        check_compatible_version(xom)
+        check_compatible_version(config)
+
 
 def test_pyramid_configure_called(makexom):
     l = []
@@ -82,7 +101,9 @@ def test_requests_only(makexom):
 
 
 @wsgi_run_throws
-def test_run_commands_called(tmpdir):
+@pytest.mark.parametrize("use_option", (False, True))
+def test_run_commands_called(tmpdir, use_option):
+    from devpi_server.init import init
     from devpi_server.main import _main, get_pluginmanager
     l = []
     class Plugin:
@@ -92,9 +113,14 @@ def test_run_commands_called(tmpdir):
             return 1
     pm = get_pluginmanager()
     pm.register(Plugin())
-    result = _main(
-        argv=["devpi-server", "--init", "--serverdir", str(tmpdir)],
-        pluginmanager=pm)
+    if use_option:
+        result = _main(
+            argv=["devpi-server", "--init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
+    else:
+        result = init(
+            argv=["devpi-init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
     result = _main(
         argv=["devpi-server", "--serverdir", str(tmpdir)],
         pluginmanager=pm)
@@ -104,7 +130,41 @@ def test_run_commands_called(tmpdir):
 
 
 @wsgi_run_throws
-def test_main_starts_server_if_run_commands_returns_none(tmpdir):
+@pytest.mark.parametrize("use_option", (False, True))
+def test_fatal_if_no_storage_and_no_sqlite_file(tmpdir, use_option):
+    from devpi_server.init import init
+    from devpi_server.main import _main, get_pluginmanager
+
+    class Plugin:
+        @hookimpl
+        def devpiserver_cmdline_run(self, xom):
+            return 1
+    pm = get_pluginmanager()
+    pm.register(Plugin())
+    if use_option:
+        _main(
+            argv=["devpi-server", "--init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
+    else:
+        init(
+            argv=["devpi-init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
+    _main(
+        argv=["devpi-server", "--serverdir", str(tmpdir)],
+        pluginmanager=pm)
+    tmpdir.join('.sqlite').remove()
+    with pytest.raises(Fatal) as excinfo:
+        _main(
+            argv=["devpi-server", "--serverdir", str(tmpdir)],
+            pluginmanager=pm
+        )
+    assert "you first need to run devpi-init or devpi-import" in str(excinfo.value)
+
+
+@wsgi_run_throws
+@pytest.mark.parametrize("use_option", (False, True))
+def test_main_starts_server_if_run_commands_returns_none(tmpdir, use_option):
+    from devpi_server.init import init
     from devpi_server.main import _main, get_pluginmanager
     l = []
     class Plugin:
@@ -113,9 +173,14 @@ def test_main_starts_server_if_run_commands_returns_none(tmpdir):
             l.append(xom)
     pm = get_pluginmanager()
     pm.register(Plugin())
-    _main(
-        argv=["devpi-server", "--init", "--serverdir", str(tmpdir)],
-        pluginmanager=pm)
+    if use_option:
+        _main(
+            argv=["devpi-server", "--init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
+    else:
+        init(
+            argv=["devpi-init", "--serverdir", str(tmpdir)],
+            pluginmanager=pm)
     with pytest.raises(ZeroDivisionError):
         _main(
             argv=["devpi-server", "--serverdir", str(tmpdir)],
@@ -126,8 +191,10 @@ def test_main_starts_server_if_run_commands_returns_none(tmpdir):
 
 def test_version_info(xom):
     app = xom.create_app()
+    while not hasattr(app, 'registry'):
+        app = app.app
     counts = {}
-    for name, version in app.app.registry['devpi_version_info']:
+    for name, version in app.registry['devpi_version_info']:
         counts[name] = counts.get(name, 0) + 1
     assert counts['devpi-server'] == 1
 
@@ -221,38 +288,50 @@ def test_no_root_pypi_option(makexom):
 
 def test_no_init_empty_directory(call_devpi_in_dir, tmpdir):
     assert not len(os.listdir(tmpdir.strpath))
-    result = call_devpi_in_dir(tmpdir, [])
+    result = call_devpi_in_dir(tmpdir, ['devpi-server'])
     assert not len(os.listdir(tmpdir.strpath))
     result.stderr.fnmatch_lines("*contains no devpi-server data*")
 
 
-def test_init_empty_directory(call_devpi_in_dir, monkeypatch, tmpdir):
+@pytest.mark.parametrize("use_option", (False, True))
+def test_init_empty_directory(call_devpi_in_dir, monkeypatch, tmpdir, use_option):
     monkeypatch.setattr("devpi_server.config.Config.init_nodeinfo", lambda x: 0/0)
     assert not len(os.listdir(tmpdir.strpath))
     with pytest.raises(ZeroDivisionError):
-        call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+        if use_option:
+            call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+        else:
+            call_devpi_in_dir(tmpdir, ["devpi-init"])
 
 
 def test_no_init_no_server_directory(call_devpi_in_dir, tmpdir):
     tmpdir.ensure("foo")
     assert os.listdir(tmpdir.strpath) == ["foo"]
-    result = call_devpi_in_dir(tmpdir, [])
+    result = call_devpi_in_dir(tmpdir, ["devpi-server"])
     assert os.listdir(tmpdir.strpath) == ["foo"]
     result.stderr.fnmatch_lines("*contains no devpi-server data*")
 
 
-def test_init_no_server_directory(call_devpi_in_dir, monkeypatch, tmpdir):
+@pytest.mark.parametrize("use_option", (False, True))
+def test_init_no_server_directory(call_devpi_in_dir, monkeypatch, tmpdir, use_option):
     monkeypatch.setattr("devpi_server.config.Config.init_nodeinfo", lambda x: 0/0)
     tmpdir.ensure("foo")
     assert os.listdir(tmpdir.strpath) == ["foo"]
     with pytest.raises(ZeroDivisionError):
-        call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+        if use_option:
+            call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+        else:
+            call_devpi_in_dir(tmpdir, ["devpi-init"])
 
 
-def test_init_server_directory(call_devpi_in_dir, tmpdir):
+@pytest.mark.parametrize("use_option", (False, True))
+def test_init_server_directory(call_devpi_in_dir, tmpdir, use_option):
     tmpdir.ensure(".nodeinfo")
     assert os.listdir(tmpdir.strpath) == [".nodeinfo"]
-    result = call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+    if use_option:
+        result = call_devpi_in_dir(tmpdir, ["devpi-server", "--init"])
+    else:
+        result = call_devpi_in_dir(tmpdir, ["devpi-init"])
     assert os.listdir(tmpdir.strpath) == [".nodeinfo"]
     result.stderr.fnmatch_lines("*already contains devpi-server data*")
 
@@ -271,3 +350,35 @@ def test_serve_max_body(monkeypatch, tmpdir):
     monkeypatch.setattr("waitress.serve", check_max_body)
     from devpi_server.main import main
     main(["devpi-server", "--max-request-body-size", "42"])
+
+
+def test_root_passwd_option(makexom):
+    # by default the password is empty
+    xom = makexom()
+    with xom.keyfs.transaction(write=False):
+        user = xom.model.get_user('root')
+        assert user.validate("")
+        assert not user.validate("foobar")
+    # the password can be set from the command line
+    xom = makexom(["--root-passwd", "foobar"])
+    with xom.keyfs.transaction(write=False):
+        user = xom.model.get_user('root')
+        assert not user.validate("")
+        assert user.validate("foobar")
+
+
+def test_root_passwd_hash_option(makexom):
+    # by default the password is empty
+    xom = makexom()
+    with xom.keyfs.transaction(write=False):
+        user = xom.model.get_user('root')
+        assert user.validate("")
+        assert not user.validate("foobar")
+    # the password hash can be directly set from the command line
+    xom = makexom(["--root-passwd-hash", "$argon2i$v=19$m=102400,t=2,p=8$j9G6V8o5B0Co9f4fQ6gVIg$WzcG2C5Bv0LwtzPWeBcz0g"])
+    # we allow writes, because the hash might be upgraded
+    # this happend after passlib 1.7.2 made argon2id the default
+    with xom.keyfs.transaction(write=True):
+        user = xom.model.get_user('root')
+        assert not user.validate("")
+        assert user.validate("foobar")
