@@ -123,7 +123,7 @@ def test_simple_project(pypistage, testapp):
     assert r.headers["X-DEVPI-SERIAL"]
     # easy_install fails if the result isn't html
     assert "html" in r.headers['content-type']
-    assert not parse_index("http://localhost", r.text, scrape=False).releaselinks
+    assert not parse_index("http://localhost", r.text).releaselinks
 
     path = "/%s-1.0.zip" % name
     pypistage.mock_simple(name, text='<a href="%s"/>' % path)
@@ -347,7 +347,21 @@ def test_simple_refresh(mapp, model, pypistage, testapp):
         info = pypistage.key_projsimplelinks("hello").get()
     assert info["links"] == []
 
-def test_inheritance_versiondata(mapp, model):
+
+def test_inheritance_project_listing_ignore_bases(mapp):
+    api1 = mapp.create_and_use()
+    mapp.upload_file_pypi("package-1.0.tar.gz", b'123',
+                          "package", "1.0", indexname=api1.stagename)
+    api2 = mapp.create_and_use(indexconfig={"bases": (api1.stagename,)})
+    mapp.upload_file_pypi("package-2.0.tar.gz", b'456',
+                          "package", "2.0", indexname=api2.stagename)
+    r = mapp.getjson(api2.index + "/package")
+    assert sorted(r["result"]) == ["1.0", "2.0"]
+    r = mapp.getjson(api2.index + "/package?ignore_bases=")
+    assert sorted(r["result"]) == ["2.0"]
+
+
+def test_inheritance_versiondata(mapp):
     api1 = mapp.create_and_use()
     mapp.upload_file_pypi("package-1.0.tar.gz", b'123',
                           "package", "1.0", indexname=api1.stagename)
@@ -428,11 +442,25 @@ def test_indexroot(testapp, model):
         user.create_stage("index", bases=("root/pypi",))
     r = testapp.get("/user/index")
     assert r.status_code == 200
+    r = testapp.get("/user/index", accept="application/json")
+    assert r.status_code == 200
+    assert "projects" in r.json["result"]
+    r = testapp.get("/user/index?no_projects", accept="application/json")
+    assert r.status_code == 200
+    assert "projects" not in r.json["result"]
+
 
 def test_indexroot_root_pypi(testapp, xom):
     r = testapp.get("/root/pypi")
     assert r.status_code == 200
     assert b"in-stage" not in r.body
+    r = testapp.get("/root/pypi", accept="application/json")
+    assert r.status_code == 200
+    assert "projects" in r.json["result"]
+    r = testapp.get("/root/pypi?no_projects=", accept="application/json")
+    assert r.status_code == 200
+    assert "projects" not in r.json["result"]
+
 
 @pytest.mark.parametrize("url", [
     '/root/pypi/{name}',
@@ -528,6 +556,7 @@ def test_apiconfig_features(testapp):
 
 def test_apiconfig_features_plugin(maketestapp, makexom):
     class Plugin:
+        @hookimpl
         def devpiserver_get_features(self):
             return set(['bar'])
     xom = makexom(plugins=[Plugin()])
@@ -540,206 +569,7 @@ def test_apiconfig_features_plugin(maketestapp, makexom):
 def test_getchanges(testapp):
     # the replica protocol should be disabled by default
     testapp.xget(403, "/+changelog/0")
-
-
-class TestStatus:
-    def test_status_master(self, testapp):
-        r = testapp.get_json("/+status", status=200)
-        assert r.status_code == 200
-        data = r.json["result"]
-        assert data["role"] == "MASTER"
-
-    def test_status_replica(self, maketestapp, replica_xom):
-        testapp = maketestapp(replica_xom)
-        r = testapp.get_json("/+status", status=200)
-        assert r.status_code == 200
-        data = r.json["result"]
-        assert data["role"] == "REPLICA"
-        assert data["serial"] == replica_xom.keyfs.get_current_serial()
-        assert data["replication-errors"] == {}
-
-
-class TestStatusInfoPlugin:
-    @pytest.fixture
-    def plugin(self):
-        from devpi_server.views import devpiweb_get_status_info
-        return devpiweb_get_status_info
-
-    def _xomrequest(self, xom):
-        from pyramid.request import Request
-        request = Request.blank("/blankpath")
-        request.registry = dict(
-            xom=xom,
-            devpi_version_info=[])
-        return request
-
-    @pytest.mark.with_notifier
-    def test_no_issue(self, plugin, xom):
-        request = self._xomrequest(xom)
-        serial = xom.keyfs.get_current_serial()
-        xom.keyfs.notifier.wait_event_serial(serial)
-        # if devpi-web is installed make sure we look again
-        # at the serial in case a hook created a new serial
-        serial = xom.keyfs.get_current_serial()
-        xom.keyfs.notifier.wait_event_serial(serial)
-        result = plugin(request)
-        assert not xom.is_replica()
-        assert result == []
-
-    @pytest.mark.xfail(reason="sometimes fail due to race condition in db table creation")
-    @pytest.mark.with_replica_thread
-    @pytest.mark.with_notifier
-    def test_no_issue_replica(self, plugin, xom):
-        request = self._xomrequest(xom)
-        serial = xom.keyfs.get_current_serial()
-        xom.keyfs.notifier.wait_event_serial(serial)
-        result = plugin(request)
-        assert hasattr(xom, 'replica_thread')
-        assert xom.is_replica()
-        assert result == []
-
-    def test_events_lagging(self, plugin, xom, monkeypatch):
-        # write transaction so event processing can lag behind
-        with xom.keyfs.transaction(write=True):
-            xom.model.create_user("hello", "pass")
-
-        import time
-        now = time.time()
-        request = self._xomrequest(xom)
-        # nothing if events never processed directly after startup
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 30)
-        result = plugin(request)
-        assert result == []
-        # fatal after 5 minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
-        result = plugin(request)
-        assert result == [dict(
-            status='fatal',
-            msg="The event processing doesn't seem to start")]
-        # fake first event processed
-        xom.keyfs.notifier.write_event_serial(0)
-        xom.keyfs.notifier.event_serial_in_sync_at = now
-        # no report in the first minute
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 30)
-        result = plugin(request)
-        assert result == []
-        # warning after 5 minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
-        result = plugin(request)
-        assert result == [dict(
-            status='warn',
-            msg='No changes processed by plugins for more than 5 minutes')]
-        # fatal after 30 minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 1810)
-        result = plugin(request)
-        assert result == [dict(
-            status='fatal',
-            msg='No changes processed by plugins for more than 30 minutes')]
-        # warning about sync after one hour
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 3610)
-        result = plugin(request)
-        assert result == [
-            dict(
-                status='warn',
-                msg="The event processing hasn't been in sync for more than 1 hour"),
-            dict(
-                status='fatal',
-                msg='No changes processed by plugins for more than 30 minutes')]
-        # fatal sync state after 6 hours
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 21610)
-        result = plugin(request)
-        assert result == [
-            dict(
-                status='fatal',
-                msg="The event processing hasn't been in sync for more than 6 hours"),
-            dict(
-                status='fatal',
-                msg='No changes processed by plugins for more than 30 minutes')]
-
-    def test_replica_lagging(self, plugin, makexom, monkeypatch):
-        from devpi_server.replica import ReplicaThread
-        import time
-        now = time.time()
-        xom = makexom(["--master=http://localhost"])
-        request = self._xomrequest(xom)
-        xom.replica_thread = ReplicaThread(xom)
-        assert xom.is_replica()
-        # fake first serial processed
-        xom.replica_thread.update_master_serial(0)
-        xom.replica_thread.replica_in_sync_at = now
-        # no report in the first minute
-        result = plugin(request)
-        assert result == []
-        # warning after one minute
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
-        result = plugin(request)
-        assert result == [dict(
-            status='warn',
-            msg='Replica is behind master for more than 1 minute')]
-        # fatal after five minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
-        result = plugin(request)
-        assert result == [dict(
-            status='fatal',
-            msg='Replica is behind master for more than 5 minutes')]
-
-    def test_initial_master_connection(self, plugin, makexom, monkeypatch):
-        from devpi_server.replica import ReplicaThread
-        import time
-        now = time.time()
-        xom = makexom(["--master=http://localhost"])
-        request = self._xomrequest(xom)
-        xom.replica_thread = ReplicaThread(xom)
-        assert xom.is_replica()
-        assert xom.replica_thread.started_at is None
-        # fake replica start
-        xom.replica_thread.started_at = now
-        # no report in the first minute
-        result = plugin(request)
-        assert result == []
-        # warning after one minute
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
-        result = plugin(request)
-        assert result == [dict(
-            status='warn',
-            msg='No contact to master for more than 1 minute')]
-        # fatal after five minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
-        result = plugin(request)
-        assert result == [dict(
-            status='fatal',
-            msg='No contact to master for more than 5 minutes')]
-
-    @pytest.mark.xfail(reason="sometimes fail due to race condition in db table creation")
-    @pytest.mark.with_replica_thread
-    @pytest.mark.with_notifier
-    def test_no_master_update(self, plugin, xom, monkeypatch):
-        import time
-        now = time.time()
-        request = self._xomrequest(xom)
-        serial = xom.keyfs.get_current_serial()
-        xom.keyfs.notifier.wait_event_serial(serial)
-        serial = xom.keyfs.get_current_serial()
-        xom.keyfs.notifier.wait_event_serial(serial)
-        assert hasattr(xom, 'replica_thread')
-        assert xom.is_replica()
-        # fake last update
-        xom.replica_thread.update_from_master_at = now
-        # no report in the first minute
-        result = plugin(request)
-        assert result == []
-        # warning after one minute
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
-        result = plugin(request)
-        assert result == [dict(
-            status='warn',
-            msg='No update from master for more than 1 minute')]
-        # fatal after five minutes
-        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
-        result = plugin(request)
-        assert result == [dict(
-            status='fatal',
-            msg='No update from master for more than 5 minutes')]
+    testapp.xget(403, "/+changelog/0-")
 
 
 def test_apiconfig_with_outside_url(testapp):
@@ -842,15 +672,30 @@ class TestSubmitValidation:
 
     def test_upload_file(self, submit, mapp):
         metadata = {"name": "Pkg5", "version": "1.0", ":action": "submit"}
-        submit.metadata(metadata, code=200)
-        r = submit.file("pkg5-2.6.tgz", b"123", {"name": "pkg5some"}, code=400)
-        assert "no project" in r.status
-        submit.file("pkg5-2.6.tgz", b"123", {"name": "Pkg5"}, code=200)
-        r = submit.file("pkg5-2.6.qwe", b"123", {"name": "Pkg5"}, code=400)
+        r = submit.file("pkg5-1.0.qwe", b"123", metadata, code=400)
         assert "not a valid" in r.status
-        r = submit.file("pkg5-2.7.tgz", b"123", {"name": "pkg5"}, code=200)
+        data = mapp.getjson("/%s" % submit.stagename)["result"]
+        assert data["projects"] == []
+        r = submit.file("pkg5-2.7.tgz", b"123", metadata, code=400)
+        assert "does not contain version" in r.status
+        data = mapp.getjson("/%s" % submit.stagename)["result"]
+        assert data["projects"] == []
+        r = submit.file("pkg5-1.0.tgz", b"123", metadata, code=200)
+        data = mapp.getjson("/%s" % submit.stagename)["result"]
+        assert data["projects"] == ["pkg5"]
         mapp.get_release_paths("Pkg5")
 
+    def test_upload_file_metadata(self, submit, mapp):
+        metadata = {"name": "Pkg5", "version": "1.0", ":action": "submit"}
+        # submit with description
+        submit.metadata(dict(metadata, description="Foo Pkg5"), code=200)
+        data = mapp.getjson("/%s/pkg5/1.0" % submit.stagename)["result"]
+        assert data["description"] == "Foo Pkg5"
+        # upload file without description
+        submit.file("pkg5-1.0.tgz", b"123", metadata, code=200)
+        data = mapp.getjson("/%s/pkg5/1.0" % submit.stagename)["result"]
+        assert data["description"] == "Foo Pkg5"
+
     def test_upload_with_removed_base(self, mapp, testapp):
         mapp.create_and_login_user("user1", "1")
         mapp.create_index("prod")
@@ -1204,6 +1049,40 @@ def test_push_from_pypi_fail(httpget, mapp, pypistage, testapp):
     assert r.json["message"] == "error 502 getting https://pypi.org/simple/hello/hello-1.0.tar.gz"
 
 
+def test_push_from_pypi_mirror_switch_to_use_external_urls(httpget, mapp, pypistage, testapp):
+    pypistage.mock_simple("hello", text='<a href="hello-1.0.tar.gz"/>')
+    pypistage.mock_extfile("/simple/hello/hello-1.0.tar.gz", b"123")
+    mapp.create_and_login_user("foo")
+    mapp.create_index("newindex1", indexconfig=dict(bases=["root/pypi"]))
+    api = mapp.use("root/pypi")
+    assert not pypistage.use_external_url
+    # download the file to the mirror index
+    pkg_url = api.simpleindex + 'hello/'
+    (tag,) = testapp.get(pkg_url).html.select('a')
+    r = testapp.get(URL(pkg_url).joinpath(tag['href']).url)
+    assert r.body == b"123"
+    with mapp.xom.keyfs.transaction(write=True):
+        # switch to external URLs
+        pypistage.modify(mirror_use_external_urls=True)
+        # and remove the file to simulate a cleanup
+        linkstore = pypistage.get_linkstore_perstage("hello", "1.0")
+        (link,) = linkstore.get_links()
+        link.entry.file_delete()
+    assert pypistage.use_external_url
+    # we should now get a redirect when trying to get the file
+    r = testapp.xget(302, URL(pkg_url).joinpath(tag['href']).url)
+    # now we try to push the release to the new index
+    req = dict(name="hello", version="1.0", targetindex="foo/newindex1")
+    r = testapp.push("/root/pypi", json.dumps(req))
+    assert r.status_code == 200
+    assert r.json == {
+        'result': [
+            [200, 'register', 'hello', '1.0', '->', 'foo/newindex1'],
+            [200, 'store_releasefile',
+             'foo/newindex1/+f/a66/5a45920422f9d/hello-1.0.tar.gz']],
+        'type': 'actionlog'}
+
+
 def test_upload_docs_for_version_without_release(mapp, testapp, monkeypatch):
     mapp.create_and_use()
     mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
@@ -1270,8 +1149,7 @@ def test_upload_and_push_internal(mapp, testapp, monkeypatch, proj):
     assert link.href.endswith("/pkg1-2.6.tgz")
 
 
-def test_acl_toxresults_upload(mapp, testapp):
-    from test_devpi_server.example import tox_result_data
+def test_acl_toxresults_upload(mapp, testapp, tox_result_data):
     mapp.create_and_login_user("user1", "1")
     mapp.create_index("prod")
     mapp.create_index("dev")
@@ -1297,8 +1175,7 @@ def test_acl_toxresults_upload(mapp, testapp):
 
 
 @pytest.mark.parametrize("outside_url", ['', 'http://localhost/devpi'])
-def test_upload_and_push_with_toxresults(mapp, testapp, outside_url):
-    from test_devpi_server.example import tox_result_data
+def test_upload_and_push_with_toxresults(mapp, testapp, outside_url, tox_result_data):
     mapp.create_and_login_user("user1", "1")
     mapp.create_index("prod")
     mapp.create_index("dev")
@@ -1343,153 +1220,6 @@ def test_upload_and_push_with_toxresults(mapp, testapp, outside_url):
     assert history_log[1]['dst'] == 'user1/prod'
 
 
-def test_upload_and_push_external(mapp, testapp, reqmock):
-    api = mapp.create_and_use()
-    mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
-    zipcontent = zip_dict({"index.html": "<html/>"})
-    mapp.upload_doc("pkg1.zip", zipcontent, "pkg1", "")
-
-    r = testapp.xget(200, api.simpleindex + "pkg1/")
-    a = getfirstlink(r.text)
-    assert "pkg1-2.6.tgz" in a.get("href")
-
-    # get root index page
-    r = testapp.xget(200, api.index)
-
-    # push OK
-    req = dict(name="pkg1", version="2.6", posturl="http://whatever.com/",
-               username="user", password="password")
-    rec = reqmock.mockresponse(url=None, code=200, method="POST", data="msg")
-    body = json.dumps(req).encode("utf-8")
-    r = testapp.request(api.index, method="POST", body=body,
-                        expect_errors=True)
-    assert r.status_code == 200
-    assert len(rec.requests) == 3
-    for i in range(3):
-        assert rec.requests[i].url == req["posturl"]
-    req = rec.requests[2]
-    # XXX properly decode www-url-encoded body and check zipcontent
-    assert b"pkg1.zip" in req.body
-    assert zipcontent in req.body
-
-    # push with error
-    reqmock.mockresponse(url=None, code=500, method="POST")
-    r = testapp.request(api.index, method="POST", body=body, expect_errors=True)
-    assert r.status_code == 502
-    result = r.json["result"]
-    assert len(result) == 1
-    assert result[0][0] == 500
-
-
-def test_upload_and_push_external_metadata12(mapp, reqmock, testapp):
-    from webob.request import cgi_FieldStorage
-    from io import BytesIO
-    api = mapp.create_and_use()
-    mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
-    mapp.set_versiondata(dict(
-        name="pkg1", version="2.6",
-        requires_python=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*"))
-    result = mapp.getjson(api.index + "/pkg1", code=200)
-    verdata = result['result']['2.6']
-    assert 'requires_python' in verdata
-    assert verdata['requires_python'] == ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*"
-    req = dict(name="pkg1", version="2.6", posturl="http://whatever.com/",
-               username="user", password="password")
-    rec = reqmock.mockresponse(url=None, code=200, method="POST", data="msg")
-    body = json.dumps(req).encode("utf-8")
-    r = testapp.request(api.index, method="POST", body=body,
-                        expect_errors=True)
-    assert r.status_code == 200
-    assert len(rec.requests) == 2
-    for i in range(2):
-        assert rec.requests[i].url == req["posturl"]
-    req = rec.requests[1]
-    fs = cgi_FieldStorage(
-        fp=BytesIO(req.body),
-        headers=req.headers,
-        environ={
-            'REQUEST_METHOD': 'POST'})
-    assert fs.getvalue('requires_python') == ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*"
-
-
-def test_upload_and_push_warehouse(mapp, testapp, reqmock):
-    # the new PyPI backend "warehouse" changes some things and they already
-    # start to affect current PyPI behaviour
-    from requests.adapters import HTTPAdapter
-    from requests.packages.urllib3.response import HTTPResponse
-    api = mapp.create_and_use()
-    mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
-    zipcontent = zip_dict({"index.html": "<html/>"})
-    mapp.upload_doc("pkg1.zip", zipcontent, "pkg1", "")
-
-    r = testapp.xget(200, api.simpleindex + "pkg1/")
-    a = getfirstlink(r.text)
-    assert "pkg1-2.6.tgz" in a.get("href")
-
-    # get root index page
-    r = testapp.xget(200, api.index)
-
-    responses = [
-        # the "register" call isn't needed anymore. All the metadata is
-        # sent together with file_upload anyway. So we get a 410 with the
-        # following reason
-        HTTPResponse(
-            body=py.io.BytesIO(b"msg"),
-            status=410, preload_content=False,
-            reason="Project pre-registration is no longer required or supported, so continue directly to uploading files."),
-        HTTPResponse(
-            body=py.io.BytesIO(b"msg"),
-            status=200, preload_content=False),
-        HTTPResponse(
-            body=py.io.BytesIO(b"msg"),
-            status=200, preload_content=False)]
-    requests = []
-
-    def process_request(self, request, kwargs):
-        response = responses.pop(0)
-        r = HTTPAdapter().build_response(request, response)
-        requests.append(request)
-        return r
-    reqmock.process_request = process_request.__get__(reqmock)
-
-    # push OK
-    req = dict(name="pkg1", version="2.6", posturl="http://whatever.com/",
-               username="user", password="password")
-    body = json.dumps(req).encode("utf-8")
-    r = testapp.request(api.index, method="POST", body=body,
-                        expect_errors=True)
-    assert r.status_code == 200
-    assert len(requests) == 3
-    for i in range(3):
-        assert requests[i].url == req["posturl"]
-    req = requests[1]
-    assert b"metadata_version" in req.body
-    assert b"sha256_digest" in req.body
-    assert b"pkg1-2.6.tgz" in req.body
-    req = requests[2]
-    assert b"metadata_version" in req.body
-    # XXX properly decode www-url-encoded body and check zipcontent
-    assert b"pkg1.zip" in req.body
-    assert zipcontent in req.body
-
-
-def test_upload_and_push_egg(mapp, testapp, reqmock):
-    api = mapp.create_and_use()
-    mapp.upload_file_pypi("pkg2-1.0-py27.egg", b"123", "pkg2", "1.0")
-    r = testapp.xget(200, api.simpleindex + "pkg2/")
-    a = getfirstlink(r.text)
-    assert "pkg2-1.0-py27.egg" in a.get("href")
-
-    # push
-    req = dict(name="pkg2", version="1.0", posturl="http://whatever.com/",
-               username="user", password="password")
-    rec = reqmock.mockresponse(url=None, data=b"msg", code=200)
-    r = testapp.push(api.index, json.dumps(req))
-    assert r.status_code == 200
-    assert len(rec.requests) == 2
-    assert rec.requests[0].url == req["posturl"]
-    assert rec.requests[1].url == req["posturl"]
-
 def test_upload_and_delete_project(mapp, testapp):
     api = mapp.create_and_use()
     mapp.delete_project("pkg1", code=404)
@@ -1601,8 +1331,7 @@ def test_upload_trigger(mapp):
         ('http://localhost', 'user1/dev', 'pkg1', '2.6')]
 
 
-def test_upload_and_testdata(mapp, testapp):
-    from test_devpi_server.example import tox_result_data
+def test_upload_and_testdata(mapp, testapp, tox_result_data):
     api = mapp.create_and_use()
     mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=200)
     path, = mapp.get_release_paths("pkg1")
@@ -1648,10 +1377,22 @@ def test_delete_version_fails_on_non_volatile(mapp):
     mapp.delete_project("pkg1/2.6", code=403)
 
 
-def test_upload_to_mirror_fails(mapp):
-    mapp.upload_file_pypi(
-            "pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=404,
-            indexname="root/pypi")
+@pytest.mark.nomocking
+def test_upload_to_mirror_fails(mapp, simpypi):
+    indexconfig = dict(
+        type="mirror",
+        mirror_url=simpypi.simpleurl,
+        mirror_cache_expiry=0,
+        volatile=True)
+    api = mapp.create_and_use(indexconfig=indexconfig)
+    name = "pkg1"
+    version = "2.6"
+    pkgver = "%s-%s.tgz" % (name, version)
+    simpypi.add_release(name, pkgver=pkgver)
+    r = mapp.upload_file_pypi(
+        pkgver, b"123", name, version, code=404,
+        indexname=api.stagename)
+    assert "cannot submit to" in r.text
 
 
 @pytest.mark.nomocking
@@ -1830,6 +1571,48 @@ def test_delete_volatile_fails(mapp):
     mapp.delete_project("pkg5", code=403)
 
 
+@pytest.mark.nomocking
+def test_mirror_use_external_urls(mapp, simpypi, testapp, xom):
+    indexconfig = dict(
+        type="mirror",
+        mirror_url=simpypi.simpleurl,
+        mirror_cache_expiry=0,
+        mirror_use_external_urls=True,
+        volatile=True)
+    api = mapp.create_and_use(indexconfig=indexconfig)
+    name = "pytest"
+    pkgver = "%s-2.6.zip" % name
+    simpypi.add_release(name, pkgver=pkgver)
+    simpypi.add_file('/%s/%s' % (name, pkgver), b'123')
+    r = testapp.get(api.index + '/+simple/%s' % name)
+    (link,) = sorted(
+        x.get('href').replace('../../', '/%s/' % api.stagename)
+        for x in getlinks(r.text))
+    path = link[1:]
+    assert '2.6' in path
+    # we should get a redirect to the original URL
+    r = testapp.xget(302, link, follow=False)
+    assert r.location == "%s/pytest/pytest-2.6.zip" % simpypi.baseurl
+    with testapp.xom.keyfs.transaction(write=False):
+        assert not getentry(testapp, path).file_exists()
+    # now turn external urls off
+    mapp.modify_index(api.stagename, indexconfig=["mirror_use_external_urls=False"])
+    # and fetch again
+    r = testapp.xget(200, link, follow=False)
+    with testapp.xom.keyfs.transaction(write=False):
+        # now the file was stored locally
+        assert getentry(testapp, path).file_exists()
+    # turn external urls on again
+    mapp.modify_index(api.stagename, indexconfig=["mirror_use_external_urls=True"])
+    # we still get the locally stored file
+    r = testapp.xget(200, link, follow=False)
+    # now remove the file, but keep metadata in place
+    with testapp.xom.keyfs.transaction(write=True):
+        getentry(testapp, path).file_delete()
+    # we should get a redirect to the original URL again
+    r = testapp.xget(302, link, follow=False)
+
+
 @pytest.mark.parametrize("volatile", [True, False])
 @pytest.mark.parametrize("restrict_modify", [None, "root"])
 def test_delete_with_acl_upload(mapp, restrict_modify, volatile, xom):
@@ -2069,6 +1852,23 @@ def test_upload_docs(mapp, testapp, proj):
     assert 'index.html' in archive.namelist()
 
 
+@proj
+def test_upload_docs_metadata(mapp, testapp, proj):
+    api = mapp.create_and_use()
+    mapp.set_versiondata(dict(
+        name="Pkg1", version="2.6",
+        description="Foo Pkg1"))
+    result = mapp.getjson("%s/pkg1/2.6" % api.index)
+    assert result["result"]["description"] == "Foo Pkg1"
+    content = zip_dict({"index.html": "<html/>"})
+    mapp.upload_doc("pkg1.zip", content, "pkg1", "2.6", code=200)
+    vv = get_view_version_links(testapp, api.index, "pkg1", "2.6", proj=proj)
+    link = vv.get_link(rel="doczip")
+    assert link.href.endswith("/pkg1-2.6.doc.zip")
+    result = mapp.getjson("%s/pkg1/2.6" % api.index)
+    assert result["result"]["description"] == "Foo Pkg1"
+
+
 def get_view_version_links(testapp, index, name, version, proj=False):
     if proj:
         url = "/".join([index, name])
